use syn::{parse_macro_input, spanned::Spanned, DeriveInput, Ident};
use proc_macro::TokenStream;
use quote::{quote, ToTokens};
use syn::{Data, Fields, Index, Type};

#[proc_macro]
pub fn derive_generic_raw(input: TokenStream) -> TokenStream {
    let ast = parse_macro_input!(input as DeriveInput);

    let data_name = ast.ident;

    let (impl_generics, ty_generics, where_clause) = ast.generics.split_for_impl();

    let repr = repr_from_data(&ast.data);

    let generalization = general_from_data(&ast.data);

    let specialization = special_from_data(&ast.data);

    quote! {
        impl #impl_generics generic::Generic for #data_name #ty_generics #where_clause {
            type Representation = #repr;

            fn generalize(self) -> Self::Representation {
                #generalization
            }
            fn specialize(rep: Self::Representation) -> Self {
                #specialization
            }
        }
    }.into()
}

fn special_from_data(data: &Data) -> proc_macro2::TokenStream {
    match data {
        Data::Struct(ref r#struct) => {
            match &r#struct.fields {
                Fields::Named(ref fields) => {
                    let bind_names: Vec<_> = fields.named.iter()
                        .map(|field| field.ident.clone().expect("Named fields must have an identifier."))
                        .collect();

                    let let_bind = generate_let_binds(bind_names.clone());

                    quote! {
                        let #let_bind = rep;
                        Self {
                            #(#bind_names ,)*
                        }
                    }
                },
                Fields::Unnamed(ref fields) => {
                    let bind_names: Vec<_> = fields.unnamed.iter()
                        .enumerate()
                        .map(|(i, field)| Ident::new(&format!("field_{i}"), field.span()))
                        .collect();

                    let let_bind = generate_let_binds(bind_names.clone());

                    quote! {
                        let #let_bind = rep;
                        Self( #(#bind_names ,)* )
                    }
                },
                Fields::Unit => quote! {
                    Self
                },
            }
        },
        Data::Enum(data_enum) => todo!(),
        Data::Union(data_union) => todo!(),
    }
}

fn generate_let_binds(mut bind_names: Vec<Ident>) -> proc_macro2::TokenStream {
    match bind_names.len() {
        0 => quote! {
            _
        },
        _ => {
            let last = bind_names.pop().unwrap();
            let last = quote! {
                generic::Field { value: #last }
            };

            bind_names.iter().fold(last, |acc, i|
                quote! {
                    generic::Product {
                        left: #i,
                        right: #acc
                    }
                }
            )
        }
    }
}

fn general_from_data(data: &Data) -> proc_macro2::TokenStream {
    match data {
        Data::Struct(ref r#struct) => {
            match &r#struct.fields {
                Fields::Named(ref fields) => {
                    let names: Vec<_> = fields.named.iter()
                        .map(|field| field.ident.as_ref()
                            .expect("Named Fields must have an identifier")
                            .to_token_stream()
                        )
                        .collect();

                    match names.len() {
                        0 => quote! {
                            let _ = self;
                            generic::Field { value: () }
                        },
                        _ => general_from_names(names)
                    }

                    
                },
                Fields::Unnamed(ref fields) => {
                    let indices: Vec<_> = fields.unnamed.iter()
                        .enumerate()
                        .map(|(i, _)| Index::from(i).to_token_stream())
                        .collect();

                    match fields.unnamed.len() {
                        0 => quote! {
                            let _ = self;
                            generic::Field { value: () }
                        },
                        _ => general_from_names(indices)
                    }
                },
                Fields::Unit => quote! { generic::Field { value: () } },
            }
        },
        Data::Enum(data_enum) => todo!(),
        Data::Union(data_union) => quote!{},
    }
}

fn general_from_names(mut names: Vec<proc_macro2::TokenStream>) -> proc_macro2::TokenStream {
    let last = names.pop().expect("Impossible");
    let expr = quote! { generic::Field { value: self.#last } };

    names.iter().fold(expr, |e, i| 
        quote! { 
            generic::Product {
                left: self.#i,
                right: #e
            }
        }
    )
}

#[proc_macro_derive(Generic)]
pub fn derive_generic(input: TokenStream) -> TokenStream {
    derive_generic_raw(input)
}

fn product_from_types(mut types: Vec<&Type>) -> proc_macro2::TokenStream {
    let last = types.pop().unwrap();
    let last = quote! { generic::Field<#last> };

    types.iter().fold(last, |e, i|
        quote! {
            generic::Product<#i, #e>
        }.into_token_stream()
    )
}

fn repr_from_data(data: &Data) -> proc_macro2::TokenStream {
    match data {
        Data::Struct(ref r#struct) => {
            match &r#struct.fields {
                Fields::Named(ref fields) => {
                    let v: Vec<_> = fields.named
                        .iter()
                        .map(|f| &f.ty)
                        .collect();

                    match v.len() {
                        0 => quote! { generic::Field<()> },
                        _ => product_from_types(v)
                    }
                },
                Fields::Unnamed(ref fields) => {
                    let v: Vec<_> = fields.unnamed
                        .iter()
                        .map(|f| &f.ty)
                        .collect();

                    match v.len() {
                        0 => quote! { generic::Field<()> },
                        _ => product_from_types(v)
                    }
                },
                Fields::Unit => quote! { generic::Field<()> },
            }
        },
        Data::Enum(data_enum) => todo!(),
        Data::Union(_) => quote! { compile_error!("Cannot derive Generics for union types!") },
    }
}
